#include <iostream>
#include <fstream>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/wifi-module.h"
#include "ns3/tap-bridge-module.h"
#include "ns3/stats-module.h"
#include "ns3/applications-module.h"

#include <iostream>
#include <string>
#include <sstream>

using namespace ns3;
//using namespace std;

NS_LOG_COMPONENT_DEFINE ("RadioWatcherExample");

/***********************************************************************************
 * 
 *      Wireless channel setup
 * 
 ***********************************************************************************/
void
setup_wifi (NodeContainer &nodes, NetDeviceContainer & devices, double sender_dbm, double interference_sensitivity, bool use_small_scale_fading) {
    WifiHelper wifi;
    YansWifiPhyHelper wifiPhy;
    YansWifiChannelHelper wifiChannel;
    NqosWifiMacHelper wifiMac;

    // Wifi 802.11g
    wifi.SetStandard (WIFI_PHY_STANDARD_80211g);
    wifiPhy = YansWifiPhyHelper::Default ();

    // Set transmission power to 11.34dbm with no antenna gains
    wifiPhy.Set ("TxPowerStart", DoubleValue (sender_dbm));
    wifiPhy.Set ("TxPowerEnd", DoubleValue (sender_dbm));
    wifiPhy.Set ("TxPowerLevels", UintegerValue (1));
    wifiPhy.Set ("EnergyDetectionThreshold", DoubleValue (-90.0));
    wifiPhy.Set ("CcaMode1Threshold", DoubleValue (interference_sensitivity));
    wifiPhy.Set ("TxGain", DoubleValue (0));
    wifiPhy.Set ("RxGain", DoubleValue (0));

    // Large scale fading -> log distance, 110m
    // Small scale fading -> Rician, K = 4.0, Coherence = 0
    wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");

    if (use_small_scale_fading)
      {
        wifiChannel.AddPropagationLoss ("ns3::LogDistancePropagationLossModel",
                "ReferenceLoss", DoubleValue (40.1157635772),
                "Exponent", DoubleValue (3.0),
                "ReferenceDistance", DoubleValue (1.0));
        wifiChannel.AddPropagationLoss ("ns3::RicePropagationLossModel",
                "K", DoubleValue (4.0),
                "Coherence", DoubleValue (0));
      } else
      {
        wifiChannel.AddPropagationLoss ("ns3::RangePropagationLossModel",
                "MaxRange", DoubleValue (62));
      }

    wifiMac = NqosWifiMacHelper::Default ();
    
    // Use MinstrelWifiManager to support rate adaption
    wifi.SetRemoteStationManager ("ns3::MinstrelWifiManager");

    // Set to adhoc mode
    wifiMac.SetType ("ns3::AdhocWifiMac");
    wifiPhy.Set ("ChannelNumber", UintegerValue (1));
    wifiPhy.SetChannel (wifiChannel.Create ());
    devices = wifi.Install (wifiPhy, wifiMac, nodes);
    wifiPhy.EnablePcapAll ("phy");
    //    wifiPhy.EnableAsciiAll ("phy");
}


int
main (int argc, char *argv[]) {
    NetDeviceContainer devices; 
    CommandLine cmd;

    //LogComponentEnable ("TapBridge", LOG_ALL);
    //LogComponentEnable ("RadioWatcherExample", LOG_ALL);
    //LogComponentEnable ("RadioWatcher", LOG_ALL);

    ns3::PacketMetadata::Enable ();
    ns3::Packet::EnablePrinting ();

    bool param_use_small_scale_fading = true;
    unsigned int param_number_of_nodes = 9;
    double param_duration = 3000.0;
    double param_interference_sensitivity = -99.0; // dBm
    double param_sender_dbm = 11.3417805547;

    cmd.Parse (argc, argv);

    /*
     * Configure simulator for emulation mode
     */
    GlobalValue::Bind ("SimulatorImplementationType", StringValue ("ns3::RealtimeSimulatorImpl"));
    GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));
    Config::SetDefault ("ns3::RealtimeSimulatorImpl::SynchronizationMode", EnumValue (RealtimeSimulatorImpl::SYNC_HARD_LIMIT));
    Config::SetDefault ("ns3::RealtimeSimulatorImpl::HardLimit", TimeValue(Seconds(0.1)));

    NodeContainer nodes;
    nodes.Create (param_number_of_nodes);

    /*
     * Set up wireless channel
     */
    setup_wifi (nodes, devices, param_sender_dbm, param_interference_sensitivity, param_use_small_scale_fading);

    /*
     * Set up topology
     */
    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator > ();
        
    // 3x3 Grid
    positionAlloc->Add (Vector (0.0, 0.0, 0.0));
    positionAlloc->Add (Vector (60.0, 0.0, 0.0));
    positionAlloc->Add (Vector (120.0, 0.0, 0.0));
    positionAlloc->Add (Vector (0.0, 60.0, 0.0));
    positionAlloc->Add (Vector (60.0, 60.0, 0.0));
    positionAlloc->Add (Vector (120.0, 60.0, 0.0));
    positionAlloc->Add (Vector (0.0, 120.0, 0.0));
    positionAlloc->Add (Vector (60.0, 120.0, 0.0));
    positionAlloc->Add (Vector (120.0, 120.0, 0.0));
    
    mobility.SetPositionAllocator (positionAlloc);
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.Install (nodes);

	//------------------------------------------------------------
	//-- Setup RadioWatcher
	//--------------------------------------------
    RadioWatcherHelper radiowatcher;
    radiowatcher.SetLocalSocketBasePort(6000);
    radiowatcher.SetLocalSocketAddress("127.0.0.1"); // optional, defaults to localhost
    radiowatcher.SetInterval(1.0); // Interval between Json messages in seconds
    ApplicationContainer rwApps;
    rwApps = radiowatcher.Install (nodes);   
    rwApps.Start(Seconds (0));
    rwApps.Stop(Seconds (param_duration - 1));

    //
    // Run the simulation
    //
    Simulator::Stop (Seconds (param_duration));
    Simulator::Run ();
    Simulator::Destroy ();
    
    //--------------------------------------
    // To see the JSON output you cann connect to the local socket using netcat:
    //     nc localhost 6000
    //--------------------------------------
  
}

