/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 FKIE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Christoph Fuchs <christoph.fuchs@fkie.fraunhofer.de>
 */

#include "radiowatcher-helper.h"

namespace ns3 {

RadioWatcherHelper::RadioWatcherHelper ()
  : m_interval (Seconds (1.0)),
    m_localSocketAddress (Ipv4Address::GetAny()),
    m_localSocketv6Address (),
    m_useIPv6 (false),
    m_localSocketBasePort (6000)
{
}

void
RadioWatcherHelper::SetLocalSocketAddress (char const *address)
{
  m_localSocketAddress = Ipv4Address(address);
   m_useIPv6 = false;
}

void
RadioWatcherHelper::SetLocalSocketv6Address (char const *address)
{
  m_localSocketv6Address = Ipv6Address(address);
  m_useIPv6 = true;
}

void
RadioWatcherHelper::SetLocalSocketBasePort (uint16_t port)
{
  m_localSocketBasePort = port;
}

void
RadioWatcherHelper::SetInterval (double interval)
{
  m_interval = Seconds(interval);
}

Ptr<RadioWatcher>
RadioWatcherHelper::EnableRadioWatcher (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd)
{
  Ptr<RadioWatcher> radiowatcher = CreateObject<RadioWatcher> ();
  radiowatcher->Init(nodeid, deviceid, nd);
  StartRadioWatcher(radiowatcher, nodeid);

  return radiowatcher;
}

Ptr<RadioWatcher>
RadioWatcherHelper::EnableRadioWatcher (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd, Mac48Address local_address)
{
  Ptr<RadioWatcher> radiowatcher = CreateObject<RadioWatcher> ();
  radiowatcher->Init(nodeid, deviceid, nd, local_address);
  StartRadioWatcher(radiowatcher, nodeid);

  return radiowatcher;
}
  
void
RadioWatcherHelper::StartRadioWatcher (Ptr<RadioWatcher> radiowatcher, uint32_t nodeid)
{
  if (m_useIPv6)
   {
     radiowatcher->StartStatServer(m_localSocketv6Address, m_localSocketBasePort + nodeid);
   }
   else
     radiowatcher->StartStatServer(m_localSocketAddress, m_localSocketBasePort + nodeid);
} 

ApplicationContainer
RadioWatcherHelper::Install (NodeContainer n)
{
  ApplicationContainer apps;
  for (NodeContainer::Iterator i = n.Begin (); i != n.End (); ++i)
    {
      Ptr<Node> node = *i;
      Ptr<NetDevice> dev = node->GetDevice (0);  // This must be the WiFiNetDevice of the node

      Ptr<RadioWatcher> radioWatcher = EnableRadioWatcher (dev->GetNode ()->GetId (), dev->GetIfIndex (), dev);
      /*
      * Alternative: Use the MAC address of the TapDevice as the local
      * MAC address
      * 
      * Get the MAC address of the TapDevice, this only works if the
      * address was manually set on the TapDevice using the SetAddress
      * method
      Ptr<NetDevice> tapdev = node->GetDevice (1);
      Ptr<TapBridge> tb = DynamicCast<TapBridge> (tapdev);
      Mac48Address local_address = Mac48Address::ConvertFrom(tb->GetAddress());
      Ptr<RadioWatcher> radioWatcher = EnableRadioWatcher (dev->GetNode ()->GetId (), dev->GetIfIndex (), dev, local_address);
      */
      
      radioWatcher->SetAttribute("Interval", TimeValue(m_interval));
      node->AddApplication (radioWatcher);
      apps.Add (radioWatcher);
    }
  return apps;
}

} // namespace ns3


