/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 Fraunhofer FKIE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Christoph Fuchs <christoph.fuchs@fkie.fraunhofer.de>
 */

#include "l2info.h"
#include "ns3/assert.h"
#include "ns3/simulator.h"
#include <unistd.h>
#include <iostream>
#include <ostream>
#include <fstream>

namespace ns3 {

LinkProvider::LinkProvider (Mac48Address localAddress)
  : m_mediaType (""),
    m_name (""),
    m_state (INACTIVE),
    m_local_linkAddress (localAddress)
{
    m_usageStat = new LinkUsageStatistics();
    m_packetStat = new PacketStatistics();
}

LinkProvider::~LinkProvider ()
{
    delete m_usageStat;
    delete m_packetStat;
}

void
LinkProvider::ResetStats ()
{
  m_usageStat->Reset();
  m_packetStat->Reset();
}
  
Json::Value
LinkProvider::ToJson()
{
  Json::Value val;
  std::ostringstream ss;
  ss << m_local_linkAddress;
  val["localLinkAddress"] = ss.str ();
  val["name"] = m_name;
  val["mediaType"] = m_mediaType;
  val["state"] = m_state;
  val["noise_level"] = m_noiseLevel;
  val["usageStat"] = m_usageStat->ToJson ();
  val["packetStat"] = m_packetStat->ToJson ();
  return val;
}


Link::Link (LinkProvider *linkProvider, Mac48Address address)
  : m_linkProvider (linkProvider),
    m_neighborAddress (address),
    m_state (INACTIVE),
    m_lastSNR (0.0),
    m_lastTxDataRate (0),
    m_lastRxDataRate (0)
  
{
    m_usageStat = new LinkUsageStatistics();
    m_packetStat = new PacketStatistics();
}

Link::~Link ()
{
    delete m_usageStat;
    delete m_packetStat;  
}

void
Link::ResetStats ()
{
  m_usageStat->Reset();
  m_packetStat->Reset();
}

Time
Link::LastActivity ()
{
  return m_packetStat->LastActivity();
}

Json::Value
Link::ToJson()
{
  Json::Value val;
  std::ostringstream ss;
  ss << m_neighborAddress;
  val["neighborAddress"] = ss.str ();
  val["lastSNR"] = m_lastSNR;
  val["lastTxDataRate"] = (u_int) m_lastTxDataRate;
  val["lastRxDataRate"] = (u_int) m_lastRxDataRate;
  val["usageStat"] = m_usageStat->ToJson ();
  val["packetStat"] = m_packetStat->ToJson ();
  return val;
}

  
PacketStatistics::PacketStatistics ()
  : m_lastActivity (Time (0)),
    m_rxPackets (0),
    m_txPackets (0),  
    m_rxFrames (0),
    m_txFrames (0),
    m_rxBit (0),
    m_txBit (0),
    m_rxFrameErrors (0),
    m_txFrameErrors (0),
    m_rxPacketErrors (0),
    m_txPacketErrors (0),
    m_collisions (0),
    m_rxLatency (0),
    m_txLatency (0),
    m_rxJitter (0),
    m_txJitter (0)
{
}

PacketStatistics::~PacketStatistics ()
{
}

void
PacketStatistics::Activity ()
{
  CriticalSection cs (mtx);
  m_lastActivity = Simulator::Now ();
}

Time
PacketStatistics::LastActivity()
{
  Time lastActivity;
  
  CriticalSection cs (mtx);
  lastActivity = m_lastActivity;
  
  return lastActivity;
}

void
PacketStatistics::Reset ()
{
  m_rxPackets = 0;
  m_txPackets = 0;
  m_rxFrames = 0;
  m_txFrames = 0;
  m_rxBit = 0;
  m_txBit = 0;
  m_rxFrameErrors = 0;
  m_txFrameErrors = 0;
  m_rxPacketErrors = 0;
  m_txPacketErrors = 0;
  m_collisions = 0;
  m_rxLatency = 0;
  m_txLatency = 0;
  m_rxJitter = 0;
  m_txJitter = 0;
}

void
PacketStatistics::FrameReceived (Ptr<const Packet> p)
{
  Activity();
  m_rxFrames++;
}

void 
PacketStatistics::FrameTransmitted (Ptr<const Packet> p)
{
  Activity();
  m_txFrames++;
}

void
PacketStatistics::FrameReceiveError (Ptr<const Packet> p)
{
  Activity();
  m_rxFrameErrors++;
}

void
PacketStatistics::FrameTransmitError ()
{
  // We do not consider transmit errors as activity as they do not
  // indicate an established link
  //Activity();
  m_txFrameErrors++;
}

void
PacketStatistics::PacketReceived (Ptr<const Packet> p)
{
  Activity();
  m_rxPackets++;
  m_rxBit += p->GetSize();  // counts bits in payload 
}

void 
PacketStatistics::PacketTransmitted (Ptr<const Packet> p)
{
  Activity();
  m_txPackets++;
  m_txBit += p->GetSize();  // counts bits in payload   
}

void 
PacketStatistics::PacketTransmitted ()
{
  Activity();
  m_txPackets++;
}

void
PacketStatistics::PacketTransmitError ()
{
  // We do not consider transmit errors as activity as they do not
  // indicate an established link
  //Activity();
  m_txPacketErrors++;
}

Json::Value
PacketStatistics::ToJson()
{
  Json::Value val;
  Time durationSinceLastActivity = Simulator::Now () - m_lastActivity;
  val["lastActivity"] = (u_int) durationSinceLastActivity.GetMicroSeconds();
  val["rxPackets"] = m_rxPackets;
  val["txPackets"] = m_txPackets;
  val["rxFrames"] = m_rxFrames;
  val["txFrames"] = m_txFrames;
  val["rxBit"] = m_rxBit;
  val["txBit"] = m_txBit;
  val["rxFrameErrors"] = m_rxFrameErrors;
  val["txFrameErrors"] = m_txFrameErrors;
  val["rxPacketErrors"] = m_rxPacketErrors;
  val["txPacketErrors"] = m_txPacketErrors;
  val["collisions"] = m_collisions;
  val["rxLatency"] = m_rxLatency;
  val["txLatency"] = m_txLatency;
  val["rxJitter"] = m_rxJitter;
  val["txJitter"] = m_txJitter;
  return val;
}


LinkUsageStatistics::LinkUsageStatistics ()
  : m_lastReset (Time(0)),
    m_durationTx (Time()),
    m_durationRx (Time()),
    m_durationCcaBusy (Time()),
    m_durationSleep (Time()),
    m_durationIdle (Time())
{
}

LinkUsageStatistics::~LinkUsageStatistics ()
{
}

double
LinkUsageStatistics::getAvgLoad (Time interval)
{
  if (interval.GetMicroSeconds() == 0)
      return 0.0;
  Time sum = m_durationTx + m_durationRx + m_durationCcaBusy; 
  return (double) sum.GetMicroSeconds() / (double) interval.GetMicroSeconds();
}

void
LinkUsageStatistics::stateChange (Time duration, enum WifiPhy::State state)
{
  switch (state)
    {
      case WifiPhy::TX:
        m_durationTx += duration;
        break;
      case WifiPhy::RX:
        m_durationRx += duration;
        break;
      case WifiPhy::CCA_BUSY:
        m_durationCcaBusy += duration;
        break;
      case WifiPhy::IDLE:
        m_durationIdle += duration;
        break;
      case WifiPhy::SLEEP:
        m_durationSleep += duration;
        break;
      default:
        break;
    }
}

void
LinkUsageStatistics::Reset ()
{
  m_lastReset = Simulator::Now ();
  m_durationTx = Seconds(0.0);
  m_durationRx = Seconds(0.0);
  m_durationCcaBusy = Seconds(0.0);
  m_durationSleep = Seconds(0.0);  
  m_durationIdle = Seconds(0.0);
}

Json::Value
LinkUsageStatistics::ToJson()
{
  Json::Value val;
  val["durationTx"] = (u_int) m_durationTx.GetMicroSeconds();
  val["durationRx"] = (u_int) m_durationRx.GetMicroSeconds();
  val["durationCcaBusy"] = (u_int) m_durationCcaBusy.GetMicroSeconds();
  val["durationSleep"] = (u_int) m_durationSleep.GetMicroSeconds();
  val["durationIdle"] = (u_int) m_durationIdle.GetMicroSeconds();
  Time interval = Simulator::Now () - m_lastReset;
  val["avgLoad"] = getAvgLoad(interval);
  val["loadInterval"] = (u_int) interval.GetMicroSeconds();
  return val;
}


L2StatServer::L2StatServer ()
  : m_listenfd (0),
    m_connfd (0),
    m_started (false)
{  
}

L2StatServer::~L2StatServer ()
{
}

void
L2StatServer::Open(u_int32_t address, u_int16_t port)
{   
    m_listenfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (m_listenfd < 0)
      {
        perror("ERROR opening socket for L2StatServer");
        exit(1);
      }
    FD_ZERO(&m_connections);
    FD_ZERO(&m_readFds);

    struct sockaddr_in m_servAddr;
    m_servAddr.sin_family = AF_INET;
    m_servAddr.sin_addr.s_addr = htonl(address);
    m_servAddr.sin_port = htons(port);
    
    int yes = 1;
    setsockopt(m_listenfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
    
    if (bind(m_listenfd, (struct sockaddr*)&m_servAddr,sizeof(m_servAddr)) < 0)
      {
        perror("ERROR binding socket for L2StatServer");
        close(m_listenfd);
        exit(1);
      }
    if(listen(m_listenfd, 10) < 0)
      {
        perror("ERROR switching socket for L2StatServer to listening mode");
        close(m_listenfd);
        exit(1);
      }
    FD_SET(m_listenfd, &m_connections);
    m_maxFd = m_listenfd;
    m_isOpen = true;
}

void
L2StatServer::Openv6(u_int8_t address[16], u_int16_t port)
{
    m_listenfd = socket(AF_INET6, SOCK_STREAM, IPPROTO_TCP);
    if (m_listenfd < 0)
      {
        perror("ERROR opening socket for L2StatServer");
        exit(1);
      }
    FD_ZERO(&m_connections);
    FD_ZERO(&m_readFds);

    struct sockaddr_in6 servAddr;

    servAddr.sin6_flowinfo = 0;
    servAddr.sin6_family = AF_INET6;
    memcpy(servAddr.sin6_addr.s6_addr, address, 16);
    //servAddr.sin6_addr = in6addr_any;
    servAddr.sin6_port = htons(port);
    
    int yes = 1;
    setsockopt(m_listenfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
    
    if (bind(m_listenfd, (struct sockaddr*)&servAddr,sizeof(servAddr)) < 0)
      {
        perror("ERROR binding socket for L2StatServer");
        close(m_listenfd);
        exit(1);
      }
    if(listen(m_listenfd, 10) < 0)
      {
        perror("ERROR switching socket for L2StatServer to listening mode");
        close(m_listenfd);
        exit(1);
      }
    FD_SET(m_listenfd, &m_connections);
    m_maxFd = m_listenfd;
    m_isOpen = true;
}

void
L2StatServer::Start()
{
  if (!m_isOpen)
    {
      perror("L2StatServer was not opened.");
      exit(1);
    }
  
  char buf[1];
  
  m_started = true;
  while (m_started)
    {
      m_readFds = m_connections;

      if (select(m_maxFd+1, &m_readFds, NULL, NULL, NULL) == -1)
        {
          perror("L2StatServer: Error in select.");
          exit(1);
        }
      
      for(int i=0; i <= m_maxFd; i++)
        {
          if (FD_ISSET(i, &m_readFds))
            {
              if (i == m_listenfd)
                {
                  m_connfd = accept(m_listenfd, NULL, NULL);
                  if(m_connfd < 0)
                    {
                      perror("ERROR at socket for L2StatServer (accept)");
                    }
                  FD_SET(m_connfd, &m_connections);
                  if (m_connfd > m_maxFd)
                   {
                     m_maxFd = m_connfd;
                   }
                }
              else
                {
                  int nbytes;
                  if ((nbytes = recv(i, buf, sizeof buf, 0)) <= 0) 
                    {
                      close(i);
                      FD_CLR(i, &m_connections);
                    }
                }
            }
        }
    }
}
  
void 
L2StatServer::Send(std::string statMessage)
{
  if (m_connfd <= 0)
    return;
  
  int len, i, j, written;
  std::string remaining;
  
  len = statMessage.length();

  for(j=0; j <= m_maxFd; j++)
    {
      if (j!=m_listenfd && FD_ISSET(j, &m_connections))
        {
          for (i=0; i<len; i += written)
            {
              remaining = statMessage.substr(i, len-i);
              written = write(j, remaining.data(), remaining.length());
              if (written == -1) break;
            }
        }
    }
}

void
L2StatServer::Close()
{
  m_started = false;
  m_isOpen = false;
  for(int i=0; i <= m_maxFd; i++)
    {
      close(i); 
    }
}

bool
L2StatServer::isOpen()
{
  return m_isOpen;
}

} // namespace ns3
