/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 Fraunhofer FKIE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Christoph Fuchs <christoph.fuchs@fkie.fraunhofer.de>
 */
#ifndef L2INFO_H
#define L2INFO_H

#include "ns3/simple-ref-count.h"
#include "ns3/packet.h"
#include "ns3/ptr.h"
#include "ns3/nstime.h"
#include "ns3/wifi-phy.h"
#include "ns3/mac48-address.h"
#include "ns3/system-mutex.h"
#include <jsoncpp/json/json.h>
#include <stdint.h>

#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>


namespace ns3 {

enum State
{
  INACTIVE = 0,
  LISTENING = 1,
  ACTIVE = 2
};

class LinkUsageStatistics;
class PacketStatistics;

/**
 * Refers to a net device like a radio or more generally a communication means.
 * Contains the device's capabilities and statistics for the device.
 *  
 */
class LinkProvider
{
public:
  LinkProvider (Mac48Address localAddress);
  ~LinkProvider ();
  void ResetStats ();  
  Json::Value ToJson();

  std::string m_mediaType;
  std::string m_name;
  State m_state;
  double m_noiseLevel; // last measured noise level in dBm. is not reset
  Mac48Address m_local_linkAddress;
  LinkUsageStatistics *m_usageStat;
  PacketStatistics *m_packetStat;
  
};

/**
 * Refers to a link on a communication means. A link is associated with
 * two communicating parties.
 * Contains the link's current state and statistics for the link.
 *  
 */
class Link : public SimpleRefCount<Link>
{
public:
  Link (LinkProvider *linkProvider, Mac48Address neighborAddress);
  ~Link ();
  void ResetStats ();
  Time LastActivity ();
  Json::Value ToJson();
  
  LinkProvider *m_linkProvider;
  Mac48Address m_neighborAddress;
  State m_state;
  double m_lastSNR;                     // is not reset
  uint64_t m_lastTxDataRate;            // in bps. is not reset
  uint64_t m_lastRxDataRate;            // in bps. is not reset
  LinkUsageStatistics *m_usageStat;
  PacketStatistics *m_packetStat;
  
};

/**
 * Data class to hold packet related statistics.
 *  
 */
class PacketStatistics
{
public:
  PacketStatistics ();
  ~PacketStatistics ();

  /**
   * Updates statistics when a frame has been received.
   * 
   */  
  void FrameReceived (Ptr<const Packet> p);

  /**
   * Updates statistics when a frame has been transmitted.
   * 
   */  
  void FrameTransmitted (Ptr<const Packet> p);
  
  /**
   * Updates statistics when a packet has been received.
   * 
   */  
  void PacketReceived (Ptr<const Packet> p);
  
  /**
   * Updates statistics when a packet has been transmitted.
   * 
   */    
  void PacketTransmitted (Ptr<const Packet> p);

  /**
   * Updates statistics when a packet has been transmitted.
   * This function only increases m_txPackets but not m_txBits.
   */    
  void PacketTransmitted ();
  
  /**
   * Updates statistics when a frame reception delivered an error.
   * 
   */    
  void FrameReceiveError (Ptr<const Packet> p);
  
  /**
   * Updates statistics when a frame transmission delivered an error.
   * 
   */    
  void FrameTransmitError ();
  
  /**
   * Updates statistics when a packet transmission delivered an error.
   * 
   */    
  void PacketTransmitError ();  
  
  /**
   * Function to call periodically to reset the statistics.
   * 
   */    
  void Reset ();
  
  /**
   * Returns a JSON representation of the packet statistics.
   * 
   */    
  Json::Value ToJson();

  /**
   * Returns the time of last activity.
   * 
   */    
  Time LastActivity();

private:

  /**
   * Tracks the time of the last activity.
   * @internal
   * 
   */  
  void Activity ();

  Time m_lastActivity;
  SystemMutex mtx;
  uint32_t m_rxPackets;
  uint32_t m_txPackets;  
  uint32_t m_rxFrames;
  uint32_t m_txFrames;
  uint32_t m_rxBit;
  uint32_t m_txBit;
  uint32_t m_rxFrameErrors;
  uint32_t m_txFrameErrors;
  uint32_t m_rxPacketErrors;
  uint32_t m_txPacketErrors;
  uint32_t m_collisions;
  uint32_t m_rxLatency;
  uint32_t m_txLatency;
  uint32_t m_rxJitter;
  uint32_t m_txJitter;
};

/**
 * Data class to hold statistics related to link usage.
 *  
 */
class LinkUsageStatistics
{
public:
  LinkUsageStatistics ();
  ~LinkUsageStatistics ();
  
  /**
   * Function to call periodically to reset the statistics.
   * 
   */      
  void Reset ();
  
  /**
   * Returns the fraction of time that the PHY was busy during the last
   * interval.
   * @param interval The interval length.
   * 
   */      
  double getAvgLoad (Time interval);
  
  /**
   * Updates statistics when the PHY changed state.
   * 
   */    
  void stateChange (Time duration, enum WifiPhy::State state);
  
  /**
   * Returns a JSON representation of the link usage statistics.
   * 
   */  
  Json::Value ToJson();

private:
  Time m_lastReset;
  /* summed up durations within current interval */
  Time m_durationTx;
  Time m_durationRx;
  Time m_durationCcaBusy;
  Time m_durationSleep;
  Time m_durationIdle;   
};


/**
 * Implements a server to broadcast the statistics date in JSON format
 * to connected clients.
 * 
 */
class L2StatServer
{
public:
  L2StatServer ();
  ~L2StatServer ();
  void Open(u_int32_t address, u_int16_t port);
  void Openv6(u_int8_t address[16], u_int16_t port);
  bool isOpen();
  void Start();
  void Close();
  void Send(std::string statMessage);
  
private:
  int m_listenfd;
  int m_connfd;
  struct sockaddr_in m_clientAddr;
  fd_set m_connections;
  fd_set m_readFds;
  int m_maxFd;
  bool m_isOpen;
  bool m_started;
};

} // namespace ns3

#endif /* L2INFO_H */
