/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 FKIE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Christoph Fuchs <christoph.fuchs@fkie.fraunhofer.de>
 */

#include "ns3/log.h"
#include "ns3/assert.h"
#include "ns3/abort.h"
#include "ns3/simulator.h"
#include "ns3/config.h"
#include "ns3/wifi-mac.h"
#include "ns3/wifi-net-device.h"
#include "radiowatcher.h"
#include <jsoncpp/json/json.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <vector>


namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("RadioWatcher");

NS_OBJECT_ENSURE_REGISTERED (RadioWatcher);

TypeId
RadioWatcher::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::RadioWatcher")
    .SetParent<Object> ()
    .AddConstructor<RadioWatcher> ()
    .AddAttribute ("Interval",
                   "Time interval between reports",
                   TimeValue (Seconds (1.0)),
                   MakeTimeAccessor (&RadioWatcher::m_interval),
                   MakeTimeChecker ())
    .AddAttribute ("LinkTimeout",
                   "Timeout for neighbor links ",
                   TimeValue (Seconds (10.0)),
                   MakeTimeAccessor (&RadioWatcher::m_link_timeout),
                   MakeTimeChecker ())
  ;
  return tid;
}

RadioWatcher::RadioWatcher ()
  : m_nodeid (0),
    m_deviceid (0),
    m_netDevice (0),
    m_statServer (),
    m_isRunning (false)
{
}

RadioWatcher::~RadioWatcher ()
{
}

void
RadioWatcher::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_isRunning = false;
  // Disconnect StatServer
  m_statServer.Close();
  m_serverThread->Join ();
  // Cleanup data structures
  delete m_linkProvider;
  m_links.clear();

  Application::DoDispose ();
}

void
RadioWatcher::StartApplication (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  std::ostringstream oss;
  oss.str ("");
  oss << "/NodeList/" << m_nodeid << "/DeviceList/" << m_deviceid;
  std::string devicepath = oss.str ();

  // ToDo: are the MAC callbacks neccessary here at all? PHY callbacks are triggered anyway.
  Config::Connect (devicepath + "/Mac/MacTx", MakeCallback (&RadioWatcher::DevTxTrace, this));
  Config::Connect (devicepath + "/Mac/MacRx", MakeCallback (&RadioWatcher::DevRxTrace, this));
  Config::Connect (devicepath + "/Mac/$ns3::RegularWifiMac/TxOkHeader", MakeCallback (&RadioWatcher::MacTxOkTrace, this));
  Config::Connect (devicepath + "/Mac/$ns3::RegularWifiMac/TxErrHeader", MakeCallback (&RadioWatcher::MacTxErrTrace, this));  

  Config::Connect (devicepath + "/RemoteStationManager/TxRtsFailed", MakeCallback (&RadioWatcher::TxRtsFailedTrace, this));
  Config::Connect (devicepath + "/RemoteStationManager/MacTxDataFailed", MakeCallback (&RadioWatcher::TxDataFailedTrace, this));
  Config::Connect (devicepath + "/RemoteStationManager/MacTxFinalRtsFailed", MakeCallback (&RadioWatcher::TxFinalRtsFailedTrace, this));
  Config::Connect (devicepath + "/RemoteStationManager/MacTxFinalDataFailed", MakeCallback (&RadioWatcher::TxFinalDataFailedTrace, this));

  Config::Connect (devicepath + "/Phy/State/RxOk", MakeCallback (&RadioWatcher::PhyRxOkTrace, this));
  Config::Connect (devicepath + "/Phy/State/RxError", MakeCallback (&RadioWatcher::PhyRxErrorTrace, this));
  Config::Connect (devicepath + "/Phy/State/Tx", MakeCallback (&RadioWatcher::PhyTxTrace, this));
  Config::Connect (devicepath + "/Phy/State/State", MakeCallback (&RadioWatcher::PhyStateTrace, this));

  Config::Connect (devicepath + "/Phy/MonitorSnifferRx", MakeCallback (&RadioWatcher::MonitorSnifferRxTrace, this));
  Config::Connect (devicepath + "/Phy/MonitorSnifferTx", MakeCallback (&RadioWatcher::MonitorSnifferTxTrace, this));
  
  m_isRunning = true;
  Simulator::ScheduleNow (&RadioWatcher::SendStatistics, this);
}

void
RadioWatcher::StopApplication (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_isRunning = false;
  std::ostringstream oss;
  oss.str ("");
  oss << "/NodeList/" << m_nodeid << "/DeviceList/" << m_deviceid;
  std::string devicepath = oss.str ();

  Config::Disconnect (devicepath + "/Mac/MacTx", MakeCallback (&RadioWatcher::DevTxTrace, this));
  Config::Disconnect (devicepath + "/Mac/MacRx", MakeCallback (&RadioWatcher::DevRxTrace, this));
  Config::Disconnect (devicepath + "/Mac/$ns3::AdhocWifiMac/TxOkHeader", MakeCallback (&RadioWatcher::MacTxOkTrace, this));
  Config::Disconnect (devicepath + "/Mac/$ns3::AdhocWifiMac/TxErrHeader", MakeCallback (&RadioWatcher::MacTxErrTrace, this));    

  Config::Disconnect (devicepath + "/RemoteStationManager/TxRtsFailed", MakeCallback (&RadioWatcher::TxRtsFailedTrace, this));
  Config::Disconnect (devicepath + "/RemoteStationManager/MacTxDataFailed", MakeCallback (&RadioWatcher::TxDataFailedTrace, this));
  Config::Disconnect (devicepath + "/RemoteStationManager/MacTxFinalRtsFailed", MakeCallback (&RadioWatcher::TxFinalRtsFailedTrace, this));
  Config::Disconnect (devicepath + "/RemoteStationManager/MacTxFinalDataFailed", MakeCallback (&RadioWatcher::TxFinalDataFailedTrace, this));

  Config::Disconnect (devicepath + "/Phy/State/RxOk", MakeCallback (&RadioWatcher::PhyRxOkTrace, this));
  Config::Disconnect (devicepath + "/Phy/State/RxError", MakeCallback (&RadioWatcher::PhyRxErrorTrace, this));
  Config::Disconnect (devicepath + "/Phy/State/Tx", MakeCallback (&RadioWatcher::PhyTxTrace, this));
  Config::Disconnect (devicepath + "/Phy/State/State", MakeCallback (&RadioWatcher::PhyStateTrace, this));
  
  Config::Disconnect (devicepath + "/Phy/MonitorSnifferRx", MakeCallback (&RadioWatcher::MonitorSnifferRxTrace, this));  
  Config::Disconnect (devicepath + "/Phy/MonitorSnifferTx", MakeCallback (&RadioWatcher::MonitorSnifferTxTrace, this));  
}
  
void
RadioWatcher::Init (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd, Mac48Address local_address)
{
  m_nodeid = nodeid;
  m_deviceid = deviceid;
  m_netDevice = nd;
  m_linkProvider = new LinkProvider(local_address);
  std::ostringstream name;
  name << "Node" << m_nodeid << "_Dev" << m_deviceid;
  m_linkProvider->m_name = name.str();
  Ptr<WifiNetDevice> wd = DynamicCast<WifiNetDevice> (nd);
  if (wd != NULL)
    {
      wd->GetMac()->SetLinkUpCallback (MakeCallback(&RadioWatcher::LinkUp, this));
      wd->GetMac()->SetLinkDownCallback (MakeCallback(&RadioWatcher::LinkDown, this));
    }
}

void
RadioWatcher::Init (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd)
{
  m_netDevice = nd;
  Init(nodeid, deviceid, nd, GetLocalDeviceAddress());
}

void
RadioWatcher::ResetStats ()
{
  m_linkProvider->ResetStats ();
  for (std::map<std::string, LinkPtr>::const_iterator it=m_links.begin(); it!=m_links.end(); ++it)
    it->second->ResetStats ();
}

void 
RadioWatcher::LinkUp (void)
{
  NS_LOG_FUNCTION (this);
  m_linkProvider->m_state = ACTIVE;
}

void 
RadioWatcher::LinkDown (void)
{
  NS_LOG_FUNCTION (this);  
  m_linkProvider->m_state = INACTIVE;  
}

void
RadioWatcher::DevTxTrace (std::string context, Ptr<const Packet> p)
{    
  NS_LOG_FUNCTION (this << "MAC tx" << m_nodeid << m_deviceid << GetLocalDeviceAddress());
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->PacketTransmitted(p);
}

void
RadioWatcher::DevRxTrace (std::string context, Ptr<const Packet> p)
{
  NS_LOG_FUNCTION (this << "MAC rx" << m_nodeid << m_deviceid << GetLocalDeviceAddress());
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->PacketReceived(p);
}

void 
RadioWatcher::MacTxOkTrace (std::string context, const WifiMacHeader &header)
{
  NS_LOG_FUNCTION (this << context << "MacTxOk " << header);
  Mac48Address dst = header.GetAddr1();
  if (dst == Mac48Address("00:00:00:00:00:00")) {
    return;
  }
  LinkPtr link = GetLink(dst);
  if (link == NULL) 
    return;
  link->m_packetStat->PacketTransmitted();
}

void 
RadioWatcher::MacTxErrTrace (std::string context, const WifiMacHeader &header)
{
  /* already covered with TxFinalDataFailedTrace */
  // std::cout << "MacTxErr " << header << "\n";
}

void
RadioWatcher::TxRtsFailedTrace (std::string context, Mac48Address address)
{
  NS_LOG_FUNCTION (this << context << address);
}

void
RadioWatcher::TxDataFailedTrace (std::string context, Mac48Address address)
{
  NS_LOG_FUNCTION (this << context << address);
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->FrameTransmitError();
  // update stats for specific Link
  LinkPtr link = GetLink(address);
  if (link == NULL) 
    return;  
  link->m_packetStat->FrameTransmitError();
}

void
RadioWatcher::TxFinalRtsFailedTrace (std::string context, Mac48Address address)
{
  NS_LOG_FUNCTION (this << context << address);
}

void
RadioWatcher::TxFinalDataFailedTrace (std::string context, Mac48Address address)
{
  NS_LOG_FUNCTION (this << context << address);
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->PacketTransmitError();
  // update stats for specific Link
  LinkPtr link = GetLink(address);
  if (link == NULL) 
    return;  
  link->m_packetStat->PacketTransmitError();
}


void
RadioWatcher::PhyRxOkTrace (std::string context, Ptr<const Packet> packet, double snr, WifiMode mode, WifiPreamble preamble)
{
  NS_LOG_FUNCTION (this << "PHY rx" << m_nodeid << m_deviceid << GetLocalDeviceAddress() << packet->GetSize() << mode << snr);
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->FrameReceived(packet);
  // update stats for specific Link
  Mac48Address source = GetPacketMacSource(packet);
  if (source == Mac48Address("00:00:00:00:00:00")) {
    return;
  }
  LinkPtr link = GetLink(source);
  if (link == NULL) 
    return;
  link->m_lastRxDataRate = mode.GetPhyRate();
  link->m_lastSNR = snr;
  link->m_packetStat->FrameReceived(packet);
}

void
RadioWatcher::PhyRxErrorTrace (std::string context, Ptr<const Packet> packet, double snr)
{
  NS_LOG_FUNCTION(this << "PHY rx Error" << m_nodeid << m_deviceid << GetLocalDeviceAddress() << snr);
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->FrameReceiveError(packet);
  // update stats for specific Link
  Mac48Address source = GetPacketMacSource(packet);
  if (source == Mac48Address("00:00:00:00:00:00")) {
    return;
  }

  LinkPtr link = GetLink(source);
  if (link == NULL) 
    return;  
  link->m_lastSNR = snr;
  link->m_packetStat->FrameReceiveError(packet);
}

void
RadioWatcher::PhyTxTrace (std::string context, Ptr<const Packet> packet, WifiMode mode, WifiPreamble preamble, uint8_t txPower)
{
  NS_LOG_FUNCTION (this << "PHY tx" << m_nodeid << m_deviceid << GetLocalDeviceAddress() << mode << txPower);
  // update stats for LinkProvider
  m_linkProvider->m_packetStat->FrameTransmitted(packet);
  // update stats for specific Link
  Mac48Address dst = GetPacketMacDestination(packet);
  if (dst == Mac48Address("00:00:00:00:00:00")) {
    return;
  }
  LinkPtr link = GetLink(dst);
  if (link == NULL) 
    return;  
  link->m_lastTxDataRate = mode.GetPhyRate();
  link->m_packetStat->FrameTransmitted(packet);
}

void
RadioWatcher::PhyStateTrace (std::string context, Time start, Time duration, enum WifiPhy::State state)
{
  NS_LOG_FUNCTION (this << "PHY state" << m_nodeid << m_deviceid << start << duration << state);
  m_linkProvider->m_usageStat->stateChange(duration, state);
}

void
RadioWatcher::MonitorSnifferRxTrace (std::string context, Ptr<const Packet> packet, uint16_t channelFreqMhz, uint16_t channelNumber, uint32_t rate, bool isShortPreamble, WifiTxVector txvector, double signalDbm, double noiseDbm)
{
  NS_LOG_FUNCTION (this << context);
  double rateInMbps = rate * 0.5;
  m_linkProvider->m_noiseLevel = noiseDbm;
  NS_LOG_DEBUG(std::cout  << m_nodeid << " MONITOR SNIFFER RX source: " << GetPacketMacSource(packet) << "rate: " << rateInMbps << " Mbps db: " << signalDbm << " Freq: " << channelFreqMhz << " Chan: " << channelNumber <<"\n");
}

void
RadioWatcher::MonitorSnifferTxTrace (std::string context, Ptr<const Packet> packet, uint16_t channelFreqMhz, uint16_t channelNumber, uint32_t rate, bool isShortPreamble, WifiTxVector txvector)
{
  NS_LOG_FUNCTION (this << context);
  double rateInMbps = rate * 0.5;
  NS_LOG_DEBUG(std::cout  << m_nodeid << " MONITOR SNIFFER TX dest: " << GetPacketMacDestination(packet) << " rate: " << rateInMbps << " Mbps power: " << txvector.GetTxPowerLevel() << " Freq: " << channelFreqMhz << " Chan: " << channelNumber << "\n");
}

Mac48Address
RadioWatcher::GetLocalDeviceAddress ()
{
  Mac48Address mac_address;
    
  Ptr<WifiNetDevice> wd = DynamicCast<WifiNetDevice> (m_netDevice);
  if (wd != NULL)
    {
      mac_address = wd->GetMac()->GetAddress();
    }

  return mac_address;
}

Mac48Address
RadioWatcher::GetPacketMacSource (Ptr<const Packet> packet)
{  
  WifiMacHeader hdr;
  Mac48Address srcAddress;

  NS_LOG_FUNCTION (this << *packet); 
  packet->PeekHeader (hdr);
  if (hdr.IsData()) 
    { 
      /*
       * AdHoc-Mode:
       * Address1: Destination
       * Address2: Source
       * Address3: BSSID
       */
       srcAddress = hdr.GetAddr2();
       NS_LOG_DEBUG (this << hdr.GetTypeString() << " " << srcAddress << " " << hdr.GetAddr1() << " " << hdr.GetAddr3() << " " << hdr.GetAddr4() << " " << packet->GetSize());  
    }
  return srcAddress;
}

Mac48Address
RadioWatcher::GetPacketMacDestination (Ptr<const Packet> packet)
{  
  WifiMacHeader hdr;
  Mac48Address dstAddress;

  NS_LOG_FUNCTION (this << *packet); 
  packet->PeekHeader (hdr);
  if (hdr.IsData()) 
    { 
      /*
       * AdHoc-Mode:
       * Address1: Destination
       * Address2: Source
       * Address3: BSSID
       */
      dstAddress = hdr.GetAddr1();
      NS_LOG_DEBUG (this << hdr.GetTypeString() << " " << dstAddress << " " << hdr.GetAddr1() << " " << hdr.GetAddr3() << " " << hdr.GetAddr4());  
    }
  return dstAddress;
}

LinkPtr
RadioWatcher::GetLink (Mac48Address peerAddress)
{
  LinkPtr linkptr;
  std::ostringstream oss;
  std::string peerAddressString;

  /* Filter out broadcast and multicast traffic as we do not
   *  attribute this to specific links
   */
  if (peerAddress.IsBroadcast() or peerAddress.IsGroup())
  {
    return NULL;
  }
  
  oss << peerAddress;  
  peerAddressString = oss.str ();
    
  if(m_links.find(peerAddressString) == m_links.end())
    { // element with key does not exist yet
      NS_LOG_DEBUG (std::cout<<"Create Link for "<< peerAddressString << std::endl);
      
      LinkPtr lp(new Link(m_linkProvider, peerAddress));
      m_links[peerAddressString] =  lp;
    }
  else
    {
      NS_LOG_DEBUG (this<<"Found Link for "<< peerAddressString << std::endl);
    }
  linkptr = m_links[peerAddressString]; 
  return linkptr;
}

void
RadioWatcher::RemoveLink (Mac48Address peerAddress)
{
  std::ostringstream oss;
  std::string peerAddressString;

  oss << peerAddress;  
  peerAddressString = oss.str ();
 
  NS_LOG_DEBUG (std::cout<<"Removing Link for "<< peerAddressString << std::endl);
  m_links.erase(peerAddressString);
}

void
RadioWatcher::RemoveLinkByKey (std::string key)
{
 
  NS_LOG_DEBUG (std::cout<<"Removing Link with key "<< key << std::endl);
  m_links.erase(key);
}

void
RadioWatcher::StartStatServer (Ipv4Address address, uint32_t port)
{
  NS_LOG_FUNCTION (this << address << port);
  m_statServer.Open(address.Get(), port);
  m_serverThread = Create<SystemThread> (
                            MakeCallback (&L2StatServer::Start, &m_statServer));
  m_serverThread->Start ();
  NS_LOG_DEBUG (this << "Server started.");
}

void
RadioWatcher::StartStatServer (Ipv6Address address, uint32_t port)
{
  u_int8_t buf[16];
  address.GetBytes(buf);
  
  NS_LOG_FUNCTION (this << address << port);
  m_statServer.Openv6(buf, port);
  m_serverThread = Create<SystemThread> (
                            MakeCallback (&L2StatServer::Start, &m_statServer));
  m_serverThread->Start ();
  NS_LOG_DEBUG (this << "Server started.");
}

void
RadioWatcher::SendStatistics ()
{
  NS_ABORT_MSG_UNLESS (this, "function called with null this pointer, now=" << Now () );
  
  if (m_statServer.isOpen())
    {
      Json::Value root;
      Json::Value links;
      root["nodeid"] = m_nodeid;
      root["deviceid"] = m_deviceid;
      root["linkProvider"] = m_linkProvider->ToJson();
      
      std::vector<std::string> links_to_delete;
      for (std::map<std::string, LinkPtr>::const_iterator it=m_links.begin(); it!=m_links.end(); ++it)
        {
          /* Send Link Info only if it has not timed out yet, otherwise remove link */
          if (Simulator::Now () - it->second->LastActivity() > m_link_timeout)
          {
            links_to_delete.push_back(it->first);
            continue;
          }
          links.append(it->second->ToJson());
        }
      /* Delete timed out links */
      for(size_t i = 0; i < links_to_delete.size(); i++)
        RemoveLinkByKey(links_to_delete[i]);
       
      root["links"] = links;  
      Json::FastWriter fastWriter;
      std::ostringstream oss;
      oss << fastWriter.write (root);
      m_statServer.Send(oss.str()); 
      Json::StyledWriter styledWriter;
      NS_LOG_DEBUG (this << styledWriter.write(root));  

      ResetStats ();
      if (m_isRunning)
        {
        Simulator::Schedule (m_interval, &RadioWatcher::SendStatistics, this);
        }
    }
    else NS_LOG_DEBUG (this << "Server not open");
}

} // namespace ns3
