/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 Fraunhofer FKIE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Christoph Fuchs <christoph.fuchs@fkie.fraunhofer.de>
 */

#ifndef RADIOWATCHER_H
#define RADIOWATCHER_H

#include "ns3/net-device.h"
#include "ns3/nstime.h"
#include "ns3/wifi-phy.h"
#include "ns3/wifi-mac-header.h"
#include "ns3/double.h"
#include "ns3/mac48-address.h"
#include "ns3/l2info.h"
#include "ns3/ptr.h"
#include "ns3/application.h"
#include "ns3/ipv4-address.h"
#include "ns3/system-thread.h"
#include <string>

namespace ns3 {

typedef Ptr<Link> LinkPtr;

/**
 * \brief RadioWatcher collects statistics for each NetDevice
 * 
 * A RadioWatcher collects layer 2 statistics for a NetDevice and
 * periodically sends these in JSON format over a local TCP socket to
 * attached listeners.
 * 
 */
class RadioWatcher  : public Application
{
public:
  static TypeId GetTypeId (void);
  RadioWatcher ();
  virtual ~RadioWatcher ();

  /**
   * Init function to be called directly after object creation
   *
   * @param nodeid the id of the node
   * @param deviceid the id of the device
   * @param nd pointer to the netdevice
   */
  void Init (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd);

  /**
   * Alternative Init function, allows to explicitly set the local
   * MAC address in the link statistics (e.g. when using a tap device)
   *
   * @param nodeid the id of the node
   * @param deviceid the id of the device
   * @param nd pointer to the netdevice
   * @param local_address local address that is given in link statistics
   */
  void Init (uint32_t nodeid, uint32_t deviceid, Ptr<NetDevice> nd, Mac48Address local_address);

  /**
   * Starts the L2StatServer on a local IPv4 address
   * to distribute statistics in JSON-Format 
   *
   * @param address local IPv4 address
   * @param port local port
   */  
  void StartStatServer (Ipv4Address address, uint32_t port);

  /**
   * Starts the L2StatServer on a local IPv6 address
   * to distribute statistics in JSON-Format 
   *
   * @param address local IPv6 address
   * @param port local port
   */  
  void StartStatServer (Ipv6Address addressv6, uint32_t port);

  /**
   * Function called when the net device transmits a packet
   *
   * @param context The configuration path of the caller.
   * @param p the packet being transmitted
   */
  void DevTxTrace (std::string context, Ptr<const Packet> p);

  /**
   * Function called when the net device receives a packet
   *
   * @param context The configuration path of the caller.
   * @param p the packet being received
   */
  void DevRxTrace (std::string context, Ptr<const Packet> p);

  void MacTxOkTrace (std::string context, const WifiMacHeader &header);
  void MacTxErrTrace (std::string context, const WifiMacHeader &header);

  /**
   * Function called when a RTS frame transmission by the considered
   * device has failed
   *
   * @param context The configuration path of the caller.
   * @param address The MAC address of the remote station
   */
  void TxRtsFailedTrace (std::string context, Mac48Address address);

  /**
   * Function called when a data frame transmission by the considered
   * device has failed
   *
   * @param context The configuration path of the caller.
   * @param address The MAC address of the remote station
   */
  void TxDataFailedTrace (std::string context, Mac48Address address);

  /**
   * Function called when the transmission of a RTS frame has
   * exceeded the retry limit
   *
   * @param context The configuration path of the caller.
   * @param address The MAC address of the remote station
   */
  void TxFinalRtsFailedTrace (std::string context, Mac48Address address);

  /**
   * Function called when the transmission of a data frame has
   * exceeded the retry limit
   *
   * @param context The configuration path of the caller.
   * @param address The MAC address of the remote station
   */
  void TxFinalDataFailedTrace (std::string context, Mac48Address address);

  /**
   * Function called when the PHY layer  of the considered
   * device receives a frame. Called by WifiStateHelper.
   *
   * @param context The configuration path of the caller.
   * @param packet The received packet.
   * @param snr The SNR of the received packet.
   * @param mode The transmission mode of the packet.
   * @param preamble The preamble of the packet.
   */
  void PhyRxOkTrace (std::string context, Ptr<const Packet> packet, double snr, WifiMode mode, WifiPreamble preamble);

  /**
   * Function called when a frame reception by the PHY
   * layer of the considered device resulted in an error due to a failure in the CRC check of
   * the frame. Called by WifiStateHelper.
   *
   * @param context The configuration path of the caller.
   * @param packet The received packet.
   * @param snr The transmission mode of the packet.
   */
  void PhyRxErrorTrace (std::string context, Ptr<const Packet> packet, double snr);

  /**
   * Function called when a frame transmission is started on the
   * PHY layer of the considered device. Called by WifiStateHelper.
   *
   * @param context The configuration path of the caller.
   * @param packet The received packet.
   * @param mode The transmission mode of the packet.
   * @param preamble The preamble of the packet.
   * @param txPower The transmit power level.
   */
  void PhyTxTrace (std::string context, Ptr<const Packet> packet, WifiMode mode, WifiPreamble preamble, uint8_t txPower);

  /**
   * Function called when the PHY layer of the considered device
   * changes state. Called by WifiStateHelper.
   * Possible states are:
   * IDLE      The PHY layer is IDLE.
   * CCA_BUSY  The PHY layer has sense the medium busy through the CCA mechanism.
   * TX        The PHY layer is sending a packet.
   * RX        The PHY layer is receiving a packet.
   * SWITCHING The PHY layer is switching to other channel.
   * SLEEP      The PHY layer is sleeping.
   *
   * @param context The configuration path of the caller.
   * @param start Time when the state started
   * @param duration Amount of time we've been in (or will be in) the state
   * @param state The state
   */
  void PhyStateTrace (std::string context, Time start, Time duration, enum WifiPhy::State state);
  
  /**
   * Function called when the PHY fires a MonitorSnifferRx trace event.
   *
   * \param packet the packet being received
   * \param channelFreqMhz the frequency in MHz at which the packet is
   *        received. Note that in real devices this is normally the
   *        frequency to which  the receiver is tuned, and this can be
   *        different than the frequency at which the packet was originally
   *        transmitted. This is because it is possible to have the receiver
   *        tuned on a given channel and still to be able to receive packets
   *        on a nearby channel.
   * \param channelNumber the channel on which the packet is received
   * \param rate the PHY data rate in units of 500kbps (i.e., the same
   *        units used both for the radiotap and for the prism header)
   * \param isShortPreamble true if short preamble is used, false otherwise
   * \param txvector the TX parameters
   * \param signalDbm signal power in dBm
   * \param noiseDbm  noise power in dBm
   */
  void MonitorSnifferRxTrace (std::string context, Ptr<const Packet> packet, uint16_t channelFreqMhz, uint16_t channelNumber, uint32_t rate, bool isShortPreamble, WifiTxVector txvector, double signalDbm, double noiseDbm);
  
  /**
   * 
   * Function called when the PHY fires a MonitorSnifferTx trace event.
   * 
   * \param packet the packet being transmitted
   * \param channelFreqMhz the frequency in MHz at which the packet is transmitted.
   * \param channelNumber the channel on which the packet is transmitted
   * \param rate the PHY data rate in units of 500kbps (i.e., the same units used both for the radiotap and for the prism header)
   * \param isShortPreamble true if short preamble is used, false otherwise
   * \param txvector the TX parameters
   * */
  void MonitorSnifferTxTrace (std::string context, Ptr<const Packet> packet, uint16_t channelFreqMhz, uint16_t channelNumber, uint32_t rate, bool isShortPreamble, WifiTxVector txvector);
  
  /**
   * 
   * Function called when the MAC tells that the link went up.
   * */
  void LinkUp (void);

  /**
   * 
   * Function called when the MAC tells that the link went down.
   * */
  void LinkDown (void);

protected:
  virtual void DoDispose(void);

private:
  virtual void StartApplication(void);
  virtual void StopApplication(void);
  /**
   * Periodically called to send statistics over L2StatServer.
   *
   */
  void SendStatistics ();
  
  /**
   * @internal
   * Returns the MAC source address of the packet (in ad hoc mode).
   */
  Mac48Address GetPacketMacSource (Ptr<const Packet> packet);

  /**
   * @internal
   * Returns the MAC destination address of the packet (in ad hoc mode).
   */
  Mac48Address GetPacketMacDestination (Ptr<const Packet> packet);

  /**
   * @internal
   * Returns the MAC address of the local device.
   */  
  Mac48Address GetLocalDeviceAddress ();

  /**
   * Returns the Link-Object for peer_address in m_links. If none exists
   * so far, a new one is created.
   * @internal
   */
  LinkPtr GetLink (Mac48Address peer_address);

  /**
   * Removes the Link-Object for peer_address in m_links if existing.
   * @internal
   */
  void
  RemoveLink (Mac48Address peerAddress);
  
  /**
   * Removes the Link-Object with corresponding key in m_links if existing.
   * @internal
   */
  void
  RemoveLinkByKey (std::string key);  
  
  /**
   * Resets the statistics for this RadioWatcher and all its Links.
   * Is called periodically each m_interval.
   * @internal
   */
  void ResetStats ();  

  uint32_t m_nodeid;
  uint32_t m_deviceid;
  Ptr<NetDevice> m_netDevice;
  L2StatServer m_statServer;
  Ptr<SystemThread> m_serverThread;

  bool m_isRunning;
  Time m_interval;
  Time m_link_timeout;
  LinkProvider *m_linkProvider;
  std::map<std::string, LinkPtr> m_links;

}; // class RadioWatcher

} // namespace ns3

#endif /* RADIOWATCHER_HELPER_H */

